/********************************************************
 *
 *  Copyright (c) 2012,  Nothing Real VFX
 *
 *          www.nothing-real.com
 *         belal@nothing-real.com
 *
 *          All rights reserved.
 *
 *  see enclosed LICENSE file for more information
 *
 ********************************************************/

#include <RixProjection.h>
#include <RixIntegrator.h>
#include <RixShadingUtils.h>

#include <string.h>
#include <fstream>

#include <ldpk/ldpk.h>

using namespace std;
using namespace ldpk;


static RtUString const US_doDistort("doDistort");
static RtUString const US_model("model");
static RtUString const US_distFile("distFile");
static RtUString const US_focal_length("focal_length");
static RtUString const US_focus_distance("focus_distance");
static RtUString const US_filmBack_W("filmBack_W");
static RtUString const US_filmBack_H("filmBack_H");
static RtUString const US_xOffset("xOffset");
static RtUString const US_yOffset("yOffset");
static RtUString const US_distortion("distortion");
static RtUString const US_anamorphicSqueeze("anamorphicSqueeze");
static RtUString const US_xCurvature("xCurvature");
static RtUString const US_yCurvature("yCurvature");
static RtUString const US_quarticDist("quarticDist");
static RtUString const US_dd2("dd2");
static RtUString const US_ud2("ud2");
static RtUString const US_vd2("vd2");
static RtUString const US_qd4("qd4");
static RtUString const US_ud4("ud4");
static RtUString const US_vd4("vd4");
static RtUString const US_d6("d6");
static RtUString const US_d8("d8");
static RtUString const US_phi("phi");
static RtUString const US_B("B");
static RtUString const US_cx02d2("cx02d2");
static RtUString const US_cy02d2("cy02d2");
static RtUString const US_cx22d2("cx22d2");
static RtUString const US_cy22d2("cy22d2");
static RtUString const US_cx04d4("cx04d4");
static RtUString const US_cy04d4("cy04d4");
static RtUString const US_cx24d4("cx24d4");
static RtUString const US_cy24d4("cy24d4");
static RtUString const US_cx44d4("cx44d4");
static RtUString const US_cy44d4("cy44d4");
static RtUString const US_rot("rot");
static RtUString const US_sqx("sqx");
static RtUString const US_sqy("sqy");
static RtUString const US_cx06d6("cx06d6");
static RtUString const US_cy06d6("cy06d6");
static RtUString const US_cx26d6("cx26d6");
static RtUString const US_cy26d6("cy26d6");
static RtUString const US_cx46d6("cx46d6");
static RtUString const US_cy46d6("cy46d6");
static RtUString const US_cx66d6("cx66d6");
static RtUString const US_cy66d6("cy66d6");

static RtUString const US_fov("fov");


class nr_tdeLens : public RixProjection
{
public:

    nr_tdeLens(
            RixContext& ctx,
            RtUString const handle,
            RixParameterList const* parms);

    ~nr_tdeLens() override;

    void RenderBegin(RixContext& ctx, RixProjectionEnvironment const& env,
                     RixParameterList const* parms) override;

    RixSCDetail GetProperty(
        ProjectionProperty property,
        void const** result) const override
    {
        switch (property)
        {
            case k_DicingHint:
                *result = &dicingHint;
                return k_RixSCUniform;
                break;
            case k_DeepMetric:
                *result = &deepMetric;
                return k_RixSCUniform;
                break;
            default:
                return k_RixSCInvalidDetail;
        }
    }

    void Project(
        RixProjectionContext &pCtx ) override;

    void distort(RtFloat *xdn, RtFloat *ydn);


private:

    DicingHint const dicingHint;
    DeepMetric const deepMetric;
    RtUString m_handle;

    plugin_loader pl;
    RtFloat lensRadius;
    RtFloat xyStep;
    RtFloat fieldOfView;
    RtFloat zoom;
    RtFloat spread;
    RtFloat maxX;
    RtFloat maxY;

    RtInt doDistort;
    RtInt model;
    RtUString distFile;
    RtFloat focus_distance;
    RtFloat focal_length;
    RtFloat filmBack_W;
    RtFloat filmBack_H;
    RtFloat fstop;
    RtFloat xOffset;
    RtFloat yOffset;
    RtFloat pixel_aspect;
    RtFloat distortion;
    RtFloat anamorphicSqueeze;
    RtFloat xCurvature;
    RtFloat yCurvature;
    RtFloat quarticDist;
    RtFloat dd2; RtFloat ud2; RtFloat vd2; RtFloat qd4; RtFloat ud4; RtFloat vd4; RtFloat d6; RtFloat d8;
    RtFloat phi; RtFloat B;
    RtFloat cx02d2; RtFloat cy02d2; RtFloat cx22d2; RtFloat cy22d2; RtFloat cx04d4; RtFloat cy04d4;
    RtFloat cx24d4; RtFloat cy24d4; RtFloat cx44d4; RtFloat cy44d4;
    RtFloat rot; RtFloat sqx; RtFloat sqy;
    RtFloat cx06d6; RtFloat cy06d6; RtFloat cx26d6; RtFloat cy26d6; RtFloat cx46d6; RtFloat cy46d6;
    RtFloat cx66d6; RtFloat cy66d6;
    RtFloat resAspect; RtFloat deviceAspect; RtFloat ratio;
    RixMessages *msg;
};


nr_tdeLens::nr_tdeLens(
    RixContext& ctx,
    RtUString const handle,
    RixParameterList const* parms)
    : dicingHint(k_Perspective),
      deepMetric( k_rayLength ),
      m_handle(handle)
{
    PIXAR_ARGUSED(ctx);
    PIXAR_ARGUSED(parms);
}

void nr_tdeLens::RenderBegin(
    RixContext& ctx,
    RixProjectionEnvironment const& env,
    RixParameterList const* parms)
{
    PIXAR_ARGUSED(ctx);

    model = 0;
    fieldOfView = 90;
    lensRadius = 5;
    xyStep = 1;
    zoom = 1;
    spread = 1;
    pixel_aspect=1;
    maxX=-1;
    maxY=1;

    // Lens User parameters (need to create for them a UI)
    doDistort = 1;
    RtUString distFile("");
    focus_distance=25;
    focal_length=35;
    filmBack_W=1.417;
    filmBack_H=0.945;
    fstop=5.f;
    xOffset=0;
    yOffset=0;
    distortion=0;
    anamorphicSqueeze=1;
    xCurvature=0;
    yCurvature=0;
    quarticDist=0;
    dd2=0; ud2=0; vd2=0; qd4=0; ud4=0; vd4=0; d6=0; d8=0;
    phi=0; B=0;
    cx02d2=0; cy02d2=0; cx22d2=0; cy22d2=0; cx04d4=0; cy04d4=0;
    cx24d4=0; cy24d4=0; cx44d4=0; cy44d4=0;
    rot=0; sqx=1; sqy=1;
    cx06d6=0; cy06d6=0; cx26d6=0; cy26d6=0; cx46d6=0; cy46d6=0;
    cx66d6=0; cy66d6=0;
    resAspect=1; deviceAspect=1; ratio=1;
    msg = (RixMessages*)ctx.GetRixInterface(k_RixMessages);

    RtInt resCnt;
    RtInt paramId;

    if ( parms->GetParamId( US_doDistort, &paramId ) == 0 )
        parms->EvalParam( paramId, 0, &doDistort );

    if ( parms->GetParamId( US_model, &paramId ) == 0 )
        parms->EvalParam( paramId, 0, &model );

    if ( parms->GetParamId( US_distFile, &paramId ) == 0 )
        parms->EvalParam( paramId, 0, &distFile );


    RixRenderState &state = *reinterpret_cast< RixRenderState * >(
        ctx.GetRixInterface( k_RixRenderState ) );
    RixRenderState::Type fmtType;

    float screenWidth = env.screenWindowRight - env.screenWindowLeft;
    float screenHeight = env.screenWindowTop - env.screenWindowBottom;
    if ( screenWidth == 0.0f || screenHeight == 0.0f )
    {
        float aspect = env.width * env.pixelAspectRatio / env.height;
        screenWidth = std::max( 2.0f, 2.0f * aspect );
        screenHeight = std::max( 2.0f, 2.0f / aspect );
    }
    float xStep = 0.25f * screenWidth / env.width;
    float yStep = 0.25f * screenHeight / env.height;
    xyStep = std::max( xStep, yStep );

    // env.deepMetric = RixProjectionEnvironment::k_rayLength;


    if ( parms->GetParamId( US_filmBack_W, &paramId ) == 0 )
    {
        parms->EvalParam( paramId, 0, &filmBack_W );
        filmBack_W = filmBack_W * 2.54f;   // film back cm
    }

    if ( parms->GetParamId( US_filmBack_H, &paramId ) == 0 )
    {
        parms->EvalParam( paramId, 0, &filmBack_H );
        filmBack_H = filmBack_H * 2.54f;   // film back cm
    }
    deviceAspect = filmBack_W/filmBack_H;

    struct { RtFloat fstop, flength, fdistance; } dof;

    const static RtUString US_DOF("Ri:DepthOfField");
    int has_dof = state.GetOption( US_DOF, &dof, 3 * sizeof( RtFloat ),  &fmtType, &resCnt );

    maxX = env.screenWindowRight;
    maxY = env.screenWindowTop;
    pixel_aspect = env.pixelAspectRatio;

    string df = distFile.CStr();
    if (df.length() > 1) // null makes some troubles inside Maya if a file was entered and deleted
                         // here, a distortion filename length/path should be greater than one character.
    {
        ifstream dFile(distFile.CStr());
        if(dFile.fail())
        {
            msg->Error("Cannot open the Distortion parameters file! Local parameters will be used instead.");
        }
        else
        {
            printf("[nr_tdeLens]: Parsing Lens Distortion file...\n");

            string param;
            string param_val;
            bool isFound;

            // get tde4_filmback_width_cm  from file
            isFound=0;
            param = "tde4_filmback_width_cm";
            param_val = "";

            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i < param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i = param.size()+1; i < temp.size(); i++)
                        param_val = param_val + temp[i];
                    filmBack_W = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);

            // get tde4_filmback_height_cm  from file
            isFound=0;
            param = "tde4_filmback_height_cm";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    filmBack_H = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_focal_length_cm  from file
            isFound=0;
            param = "tde4_focal_length_cm";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    focal_length = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_pixel_aspect  from file
            isFound=0;
            param = "tde4_pixel_aspect";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    pixel_aspect = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_custom_focus_distance_cm  from file
            isFound=0;
            param = "tde4_custom_focus_distance_cm";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    focus_distance = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_x_cm  from file
            isFound=0;
            param = "tde4_lens_center_offset_x_cm";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    xOffset = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "tde4_lens_center_offset_y_cm";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    yOffset = strtod(param_val.c_str(), NULL);
                    break;
                }
            }


            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Distortion";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    distortion = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Anamorphic Squeeze";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    anamorphicSqueeze = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Curvature X";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    xCurvature = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Curvature Y";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    yCurvature = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Quartic Distortion";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    quarticDist = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Distortion - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    dd2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "U - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    ud2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "V - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    vd2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Quartic Distortion - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    qd4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "U - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    ud4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "V - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    vd4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Phi - Cylindric Direction";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    phi = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "B - Cylindric Bending";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    B = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Lens Rotation";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    rot = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Squeeze-X";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    sqx = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Squeeze-Y";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    sqy = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx02 - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx02d2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy02 - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy02d2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx22 - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx22d2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy22 - Degree 2";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy22d2 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx04 - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx04d4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy04 - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy04d4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx24 - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx24d4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy24 - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy24d4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx44 - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx44d4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy44 - Degree 4";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy44d4 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx06 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx06d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy06 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy06d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx26 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx26d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy26 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy26d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx46 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx46d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy46 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy46d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cx66 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cx66d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }

            dFile.clear();
            dFile.seekg (0, ios::beg);
            // get tde4_lens_center_offset_y_cm  from file
            isFound=0;
            param = "Cy66 - Degree 6";
            param_val = "";
            while(!dFile.eof())
            {
                string temp = "";
                getline(dFile, temp);
                for(unsigned int i=0; i<param.size(); i++)
                {
                    if(temp[i] == param[i])
                        isFound = 1;
                    else
                    {
                        isFound =0;
                        break;
                    }
                }
                if(isFound)
                {
                    for(unsigned long i=param.size()+1; i<temp.size(); i++)
                        param_val = param_val + temp[i];
                    cy66d6 = strtod(param_val.c_str(), NULL);
                    break;
                }
            }


        }
    }
    else
    {
        // in case DOF is disabled, both focal_length and focus_distance will be 0
        // and user in this case should
        // so, assume they are given as params by the user in CM
        if (!has_dof) // no Ri Options for dof was available
        {
            printf("[nr_tdeLens]: DOF is not not enabled, the focal length will be used from the user parameters\n");
            if ( parms->GetParamId( US_focal_length, &paramId ) == 0 )
            {
                parms->EvalParam( paramId, 0, &focal_length );
                focal_length = (focal_length / 10.0f);	// convert to cm (applications are using mm in the user parameters of cameras)

                if ( parms->GetParamId( US_focus_distance, &paramId ) == 0 )
                    parms->EvalParam( paramId, 0, &focus_distance );  // should be given in cm

                fstop = 5.f;
            }
        }
        else
        {
            // get focal_length and focus_distance from DOF if dof is enabled
            focal_length = dof.flength; 			// Already in cm
            focus_distance = dof.fdistance;			// Already in cm
            fstop = dof.fstop;
        }

        if ( parms->GetParamId( US_xOffset, &paramId ) == 0 )
        {
            parms->EvalParam( paramId, 0, &xOffset );
            xOffset = xOffset * 2.54f;
        }

        if ( parms->GetParamId( US_yOffset, &paramId ) == 0 )
        {
            parms->EvalParam( paramId, 0, &yOffset );
            yOffset = yOffset * 2.54f;
        }

        if ( parms->GetParamId( US_distortion, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &distortion );

        if ( parms->GetParamId( US_anamorphicSqueeze, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &anamorphicSqueeze );

        if ( parms->GetParamId( US_xCurvature, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &xCurvature );

        if ( parms->GetParamId( US_yCurvature, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &yCurvature );

        if ( parms->GetParamId( US_quarticDist, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &quarticDist );

        if ( parms->GetParamId( US_dd2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &dd2 );

        if ( parms->GetParamId( US_ud2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &ud2 );

        if ( parms->GetParamId( US_vd2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &vd2 );

        if ( parms->GetParamId( US_qd4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &qd4 );

        if ( parms->GetParamId( US_ud4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &ud4 );

        if ( parms->GetParamId( US_vd4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &vd4 );

        if ( parms->GetParamId( US_d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &d6 );

        if ( parms->GetParamId( US_d8, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &d8 );

        if ( parms->GetParamId( US_phi, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &phi );

        if ( parms->GetParamId( US_B, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &B );

        if ( parms->GetParamId( US_cx02d2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx02d2 );

        if ( parms->GetParamId( US_cy02d2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy02d2 );

        if ( parms->GetParamId( US_cx22d2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx22d2 );

        if ( parms->GetParamId( US_cy22d2, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy22d2 );

        if ( parms->GetParamId( US_cx04d4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx04d4 );

        if ( parms->GetParamId( US_cy04d4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy04d4 );

        if ( parms->GetParamId( US_cx24d4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx24d4 );

        if ( parms->GetParamId( US_cy24d4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy24d4 );

        if ( parms->GetParamId( US_cx44d4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx44d4 );

        if ( parms->GetParamId( US_cy44d4, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy44d4 );

        if ( parms->GetParamId( US_rot, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &rot );

        if ( parms->GetParamId( US_sqx, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &sqx );

        if ( parms->GetParamId( US_sqy, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &sqy );

        if ( parms->GetParamId( US_cx06d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx06d6 );

        if ( parms->GetParamId( US_cy06d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy06d6 );

        if ( parms->GetParamId( US_cx26d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx26d6 );

        if ( parms->GetParamId( US_cy26d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy26d6 );

        if ( parms->GetParamId( US_cx46d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx46d6 );

        if ( parms->GetParamId( US_cy46d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy46d6 );

        if ( parms->GetParamId( US_cx66d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cx66d6 );

        if ( parms->GetParamId( US_cy66d6, &paramId ) == 0 )
            parms->EvalParam( paramId, 0, &cy66d6 );
    }

    // ------------------ END PARAM EVAL ----------------------- //
    // Collecting the lens parameters is now DONE (it was queried above from the direct lens params or from the distFile)

    fieldOfView = 2.0f * atan(filmBack_W / (2.0f * focal_length));
    fieldOfView *= F_RADTODEG;
    zoom = tanf( 0.5f * fieldOfView * F_DEGTORAD );
    spread = atanf( zoom * xyStep );
    lensRadius = 0.5f * focal_length / fstop;


	//cout << "focal length: " << focal_length << endl;
	//cout << "focus dist:   " << focus_distance << endl;
	//cout << "fstop:        " << fstop << endl;
	//cout << "filmBack_W:   " << filmBack_W << endl;
	//cout << "filmBack_H:   " << filmBack_H << endl;
	//cout << "fieldOfView:  " << fieldOfView << endl;
	//cout << "zoom:         " << zoom << endl;
	//cout << "spread:       " << spread << endl;
	//cout << "lensRadius:   " << lensRadius << endl;


    if (doDistort)
    {
        string modelPath;
        string envLens;

        const char *testEnv = getenv("LENSMODEL_PATH");
        if (testEnv == nullptr)
                msg->Error("LENSMODEL_PATH was not set!! Set LENSMODEL_PATH to ldpk plugin models path.");

        envLens = testEnv;

        if (model == 0)
        {
            modelPath = envLens + "/tde4_ldp_classic_3de_mixed" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                msg->Error("tde4_ldp_classic_3de_mixed.so is not exist! Make sure you set env LENSMODEL_PATH");
            }
        }
        else if (model == 1)
        {
            modelPath = envLens + "/tde4_ldp_radial_decentered_deg_4_cylindric" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
                msg->Error("tde4_ldp_radial_decentered_deg_4_cylindric.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else if (model == 2)
        {
            modelPath = envLens + "/tde4_ldp_anamorphic_deg_4_rotate_squeeze_xy" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
                msg->Error("tde4_ldp_anamorphic_deg_4_rotate_squeeze_xy.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else if (model == 3)
        {
            modelPath = envLens + "/tde4_ldp_radial_deg_8" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
                msg->Error("tde4_ldp_radial_deg_8.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else if (model == 4)
        {
            modelPath = envLens + "/tde4_ldp_anamorphic_deg_6" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
                msg->Error("tde4_ldp_anamorphic_deg_6.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else // Disable distortion if an invalid model number was given
            msg->Error("The distortion model given is invalid!!");

        model_parser mp(false);
        pl.open_plugin(modelPath);

        // Set the parameters that are common for any lens
        pl.get_model()->setParameterValue("tde4_focal_length_cm", focal_length);
        pl.get_model()->setParameterValue("tde4_filmback_width_cm", filmBack_W);
        pl.get_model()->setParameterValue("tde4_filmback_height_cm", filmBack_H);
        pl.get_model()->setParameterValue("tde4_pixel_aspect", pixel_aspect);
        pl.get_model()->setParameterValue("tde4_lens_center_offset_x_cm", xOffset);
        pl.get_model()->setParameterValue("tde4_lens_center_offset_y_cm", yOffset);
        pl.get_model()->setParameterValue("tde4_custom_focus_distance_cm", focus_distance);

        switch (model)
        {
        case 0:
            // Set the paramters that are specific to the 'Classic 3DE distortion' model
            pl.get_model()->setParameterValue("Distortion" , distortion);
            pl.get_model()->setParameterValue("Anamorphic Squeeze", anamorphicSqueeze);
            pl.get_model()->setParameterValue("Curvature X", xCurvature);
            pl.get_model()->setParameterValue("Curvature Y", yCurvature);
            pl.get_model()->setParameterValue("Quartic Distortion", quarticDist);
            break;
        case 1:
            // Set the paramters that are specific to the 'Radial - Standard, Degree 4' model
            pl.get_model()->setParameterValue("Distortion - Degree 2" , dd2);
            pl.get_model()->setParameterValue("U - Degree 2" , ud2);
            pl.get_model()->setParameterValue("V - Degree 2" , vd2);
            pl.get_model()->setParameterValue("Quartic Distortion - Degree 4" , qd4);
            pl.get_model()->setParameterValue("U - Degree 4" , ud4);
            pl.get_model()->setParameterValue("V - Degree 4" , vd4);
            pl.get_model()->setParameterValue("Phi - Cylindric Direction" , phi);
            pl.get_model()->setParameterValue("B - Cylindric Bending" , B);
            break;
        case 2:
            // Set the paramters that are specific to the 'Anamorphic - Standard, Degree 4' model
            pl.get_model()->setParameterValue("Cx02 - Degree 2", cx02d2);
            pl.get_model()->setParameterValue("Cy02 - Degree 2", cy02d2);
            pl.get_model()->setParameterValue("Cx22 - Degree 2", cx22d2);
            pl.get_model()->setParameterValue("Cy22 - Degree 2", cy22d2);
            pl.get_model()->setParameterValue("Cx04 - Degree 4", cx04d4);
            pl.get_model()->setParameterValue("Cy04 - Degree 4", cy04d4);
            pl.get_model()->setParameterValue("Cx24 - Degree 4", cx24d4);
            pl.get_model()->setParameterValue("Cy24 - Degree 4", cy24d4);
            pl.get_model()->setParameterValue("Cx44 - Degree 4", cx44d4);
            pl.get_model()->setParameterValue("Cy44 - Degree 4", cy44d4);
            pl.get_model()->setParameterValue("Lens Rotation", rot);
            pl.get_model()->setParameterValue("Squeeze-X", sqx);
            pl.get_model()->setParameterValue("Squeeze-Y", sqy);
            break;
        case 3:
            // Set the paramters that are specific to the 'Radial - Fisheye, Degree 8' model
            pl.get_model()->setParameterValue("Distortion - Degree 2", dd2);
            pl.get_model()->setParameterValue("Quartic Distortion - Degree 4", qd4);
            pl.get_model()->setParameterValue("Degree 6" , d6);
            pl.get_model()->setParameterValue("Degree 8" , d8);
            break;
        case 4:
            pl.get_model()->setParameterValue("Cx02 - Degree 2", cx02d2);
            pl.get_model()->setParameterValue("Cy02 - Degree 2", cy02d2);
            pl.get_model()->setParameterValue("Cx22 - Degree 2", cx22d2);
            pl.get_model()->setParameterValue("Cy22 - Degree 2", cy22d2);
            pl.get_model()->setParameterValue("Cx04 - Degree 4", cx04d4);
            pl.get_model()->setParameterValue("Cy04 - Degree 4", cy04d4);
            pl.get_model()->setParameterValue("Cx24 - Degree 4", cx24d4);
            pl.get_model()->setParameterValue("Cy24 - Degree 4", cy24d4);
            pl.get_model()->setParameterValue("Cx44 - Degree 4", cx44d4);
            pl.get_model()->setParameterValue("Cy44 - Degree 4", cy44d4);
            pl.get_model()->setParameterValue("Cx06 - Degree 6", cx06d6);
            pl.get_model()->setParameterValue("Cy06 - Degree 6", cy06d6);
            pl.get_model()->setParameterValue("Cx26 - Degree 6", cx26d6);
            pl.get_model()->setParameterValue("Cy26 - Degree 6", cy26d6);
            pl.get_model()->setParameterValue("Cx46 - Degree 6", cx46d6);
            pl.get_model()->setParameterValue("Cy46 - Degree 6", cy46d6);
            pl.get_model()->setParameterValue("Cx66 - Degree 6", cx66d6);
            pl.get_model()->setParameterValue("Cy66 - Degree 6", cy66d6);
            break;
        }

        pl.get_model()->initializeParameters();

        // Parsing Distortion Parameters from file if one exists,
        // This overrides any local parameters.
        // I found this useful in many cases and makes consistent workflow between Arnold and 3DE
        // Also the one could use it as a lens Specs file for other non-tracked scenes.
        //const char * distFile = distFile;
        if (df.length() > 1) // null makes some troubles inside Maya if a file was entered and deleted
                             // here, a distortion filename length/path should be greater than one character.
        {
            ifstream dFile(distFile.CStr());
            if(dFile.fail())
            {
                msg->Error("Cannot open the Distortion parameters file! Local parameters will be used instead.");
            }
            else
            {
                // We get the Common Lens parameters from local parameters if were not been set
                // in the param file.
            	mp.set_model(pl.get_model());
            	// FIXME: The following line is crashing the render!
                mp.parse(dFile);
            }
        }
    }
}


nr_tdeLens::~nr_tdeLens()
{
}

void nr_tdeLens::distort(RtFloat *xdn, RtFloat *ydn)
{
    double x_dn0, y_dn0;
    double x_dn = *xdn;
    double y_dn = *ydn;

    pl.get_model()->undistort(x_dn, y_dn, x_dn0, y_dn0);

    *xdn = x_dn0;
    *ydn = y_dn0;
}


void nr_tdeLens::Project(
    RixProjectionContext& pCtx )
{
    for ( int index = 0; index < pCtx.numRays; ++index )
    {
        RtPoint2 scn;
        RtPoint2 const &screen( pCtx.screen[ index ] );
        RtPoint2 const &aperture( pCtx.aperture[ index ] );
        RtRayGeometry &ray( pCtx.rays[ index ] );

        scn = screen;

        if (doDistort)
        {
            // normalize screen window from (-min to +max) to (0 to 1)
            scn.x = (scn.x + maxX) / (2.0f * maxX);
            scn.y = (scn.y + maxY) / (2.0f * maxY);

            distort(&scn.x, &scn.y);

            // inverse the normalization function to reformat pdn form -min to +max
            scn.x = (scn.x - 0.5f) * 2.0f * maxX;
            scn.y = (scn.y - 0.5f) * 2.0f * maxY;

            scn.x += (xOffset);
            scn.y += (yOffset);
        }

        ray.origin = RtPoint3(
            aperture.x * lensRadius,
            aperture.y * lensRadius,
            0.0f );

        ray.direction = RtVector3(
            scn.x * zoom * focus_distance - ray.origin.x,
            scn.y * zoom * focus_distance - ray.origin.y,
            focus_distance );

        ray.direction.Normalize();
        ray.originRadius = 0.0f;
        ray.raySpread = spread * ray.direction.z;
    }
}


// ===============================================================

class nr_tdeLensFactory : public RixProjectionFactory
{
public:

    nr_tdeLensFactory() {}

    int Init(RixContext& ctx, RtUString const pluginPath) override
    {
        PIXAR_ARGUSED(ctx);
        PIXAR_ARGUSED(pluginPath);

        return 0;
    }

    void Finalize(RixContext& ctx) override
    {
        PIXAR_ARGUSED(ctx);
    }

    RixSCParamInfo const* GetParamTable() override;

    void Synchronize(
        RixContext&,
        RixSCSyncMsg,
        RixParameterList const*) override
    {
    }

    RixProjection* CreateProjection(
        RixContext& ctx,
        RtUString const handle,
        RixParameterList const* pList) override;

    void DestroyProjection(RixProjection const* projection) override;
};



RixSCParamInfo const* nr_tdeLensFactory::GetParamTable()
{
    static RixSCParamInfo s_ptable[] =
    {
        RixSCParamInfo( US_doDistort, k_RixSCInteger ),
        RixSCParamInfo( US_model, k_RixSCInteger ),
        RixSCParamInfo( US_distFile, k_RixSCString ),
        RixSCParamInfo( US_focal_length, k_RixSCFloat ),
        RixSCParamInfo( US_focus_distance, k_RixSCFloat ),
        RixSCParamInfo( US_filmBack_W, k_RixSCFloat ),
        RixSCParamInfo( US_filmBack_H, k_RixSCFloat ),
        RixSCParamInfo( US_xOffset, k_RixSCFloat ),
        RixSCParamInfo( US_yOffset, k_RixSCFloat ),
        RixSCParamInfo( US_distortion, k_RixSCFloat ),
        RixSCParamInfo( US_anamorphicSqueeze, k_RixSCFloat ),
        RixSCParamInfo( US_xCurvature, k_RixSCFloat ),
        RixSCParamInfo( US_yCurvature, k_RixSCFloat ),
        RixSCParamInfo( US_quarticDist, k_RixSCFloat ),
        RixSCParamInfo( US_dd2, k_RixSCFloat ),
        RixSCParamInfo( US_ud2, k_RixSCFloat ),
        RixSCParamInfo( US_vd2, k_RixSCFloat ),
        RixSCParamInfo( US_qd4, k_RixSCFloat ),
        RixSCParamInfo( US_ud4, k_RixSCFloat ),
        RixSCParamInfo( US_vd4, k_RixSCFloat ),
        RixSCParamInfo( US_d6, k_RixSCFloat ),
        RixSCParamInfo( US_d8, k_RixSCFloat ),
        RixSCParamInfo( US_phi, k_RixSCFloat ),
        RixSCParamInfo( US_B, k_RixSCFloat ),
        RixSCParamInfo( US_cx02d2, k_RixSCFloat ),
        RixSCParamInfo( US_cy02d2, k_RixSCFloat ),
        RixSCParamInfo( US_cx22d2, k_RixSCFloat ),
        RixSCParamInfo( US_cy22d2, k_RixSCFloat ),
        RixSCParamInfo( US_cx04d4, k_RixSCFloat ),
        RixSCParamInfo( US_cy04d4, k_RixSCFloat ),
        RixSCParamInfo( US_cx24d4, k_RixSCFloat ),
        RixSCParamInfo( US_cy24d4, k_RixSCFloat ),
        RixSCParamInfo( US_cx44d4, k_RixSCFloat ),
        RixSCParamInfo( US_cy44d4, k_RixSCFloat ),
        RixSCParamInfo( US_rot, k_RixSCFloat ),
        RixSCParamInfo( US_sqx, k_RixSCFloat ),
        RixSCParamInfo( US_sqy, k_RixSCFloat ),
        RixSCParamInfo( US_cx06d6, k_RixSCFloat ),
        RixSCParamInfo( US_cy06d6, k_RixSCFloat ),
        RixSCParamInfo( US_cx26d6, k_RixSCFloat ),
        RixSCParamInfo( US_cy26d6, k_RixSCFloat ),
        RixSCParamInfo( US_cx46d6, k_RixSCFloat ),
        RixSCParamInfo( US_cy46d6, k_RixSCFloat ),
        RixSCParamInfo( US_cx66d6, k_RixSCFloat ),
        RixSCParamInfo( US_cy66d6, k_RixSCFloat ),

        RixSCParamInfo( US_fov, k_RixSCFloat ),

        RixSCParamInfo() // end of table
    };
    return &s_ptable[ 0 ];
}

RixProjection* nr_tdeLensFactory::CreateProjection(
    RixContext& ctx,
    RtUString const handle,
    RixParameterList const* pList)
{
    return new nr_tdeLens(ctx, handle, pList);
}

void nr_tdeLensFactory::DestroyProjection(RixProjection const* projection)
{
    delete (nr_tdeLens*)projection;
}


RIX_PROJECTIONFACTORYCREATE
{
    PIXAR_ARGUSED(hint);
    return new nr_tdeLensFactory();
}

RIX_PROJECTIONFACTORYDESTROY
{
    delete reinterpret_cast< nr_tdeLensFactory * >( factory );
}
