﻿#
#
# 3DE4.script.name:	SItoA bs_tdeLens...
#
# 3DE4.script.version:	v1.0
#
# 3DE4.script.gui:	Main Window::3DE4::Export Project
#
# 3DE4.script.comment:	Creates a vb script file that contains all project data, which can be imported into Softimage for SItoA bs_tdeLens.
#
#

#
# import sdv's python vector lib...

from vl_sdv import *

#
# functions...

def convertToAngles(r3d):
	rot	= rot3d(mat3d(r3d)).angles(VL_APPLY_ZXY)
	rx	= (rot[0]*180.0)/3.141592654
	ry	= (rot[1]*180.0)/3.141592654
	rz	= (rot[2]*180.0)/3.141592654
	return(rx,ry,rz)


def convertZup(p3d,yup):
	if yup==1:
		return(p3d)
	else:
		return([p3d[0],-p3d[2],p3d[1]])


def angleMod360(d0,d):
	dd	= d-d0
	if dd>180.0:
		d	= angleMod360(d0,d-360.0)
	else:
		if dd<-180.0:
			d	= angleMod360(d0,d+360.0)
	return d


def validName(name):
	name	= name.replace(" ","_")
	numbers = ["0","1","2","3","4","5","6","7","8","9"]
	if name[0] in numbers:
		name = "_" + name
	return name



#
# main script...
		

#
# search for camera point group...

campg	= None
pgl	= tde4.getPGroupList()
for pg in pgl:
	if tde4.getPGroupType(pg)=="CAMERA": campg = pg
if campg==None:
	tde4.postQuestionRequester("Export Softimage XSI...","Error, there is no camera point group.","Ok")


#
# open requester...

req	= tde4.createCustomRequester()
tde4.addFileWidget(req,"file_browser","Exportfile...","*.vbs")
tde4.addTextFieldWidget(req, "startframe_field", "Startframe", "1")
ret	= tde4.postCustomRequester(req,"Export Softimage XSI (VB-Script)...",540,120,"Ok","Cancel")
if ret==1:
	yup	= 1
#	if yup==2: yup = 0
	path	= tde4.getWidgetValue(req,"file_browser")
	frame0	= float(tde4.getWidgetValue(req,"startframe_field"))
	frame0	-= 1
	if path!=None:
		f	= open(path,"w")
		if not f.closed:
			
			#
			# write some comments...
			
			f.write("''\n")
			f.write("'' Softimage XSI/VB export data written by %s\n"%tde4.get3DEVersion())
			f.write("''\n")
			f.write("'' All lengths are in centimeter, all angles are in degree.\n")
			f.write("''\n\n")
			

			f.write("'' Switch off command logging for speed improvement\n")
			f.write("SetValue \"preferences.scripting.cmdlog\", False\n")


			#
			# write scene group...
			
			f.write("'' create scene group...\n")
			f.write("set sceneGroupName = GetPrim(\"Null\",\"Scene\")\n")
			f.write("SetValue sceneGroupName & \".null.primary_icon\", 0\n")
			
			#
			# write cameras...
			
			cl	= tde4.getCameraList()
			index	= 1
			for cam in cl:
				noframes	= tde4.getCameraNoFrames(cam)
				lens		= tde4.getCameraLens(cam)
				cattr = tde4.getCameraSequenceAttr(cam)
				filename = tde4.getCameraPath(cam)
				padding = 0
				while(filename.count("##") != 0):
					filename = filename.replace("##","#",1)
					padding += 1
				padding += 1
				seqname = filename.replace("#","[%d..%d;%d]"%(cattr[0],cattr[1],padding))
				clipname = validName(seqname)
				if lens!=None:
					name		= validName(tde4.getCameraName(cam))
					name		= "%s_%s_1"%(name,index)
					index		+= 1
					fback_w		= tde4.getLensFBackWidth(lens)
					fback_h		= tde4.getLensFBackHeight(lens)
					p_aspect	= tde4.getLensPixelAspect(lens)
					f_aspect	= tde4.getLensFilmAspect(lens)
					focal		= tde4.getCameraFocalLength(cam,1)
					prjploffx	= (-1)*tde4.getLensLensCenterX(lens)/2.54
					prjploffy	= (-1)*tde4.getLensLensCenterY(lens)/2.54
					
					# convert filmback to inch...
					fback_w		= fback_w/2.54
					fback_h		= fback_h/2.54
					
					# convert focal length to mm...
					focal		= focal*10.0
					camera_rig = "%s_rig"%(name)
					# create camera...
					f.write("\n")
					f.write("'' create camera %s...\n"%camera_rig)
					f.write("set cam = sceneGroupName.AddCameraRig(\"Camera\")\n")
					
					#change camera name
					f.write("SetValue cam & \".Name\", \"%s\"\n"%camera_rig)
					f.write("set camera = cam.camera\n")
					f.write("SetValue cam & \".camera1.Name\", \"%s\"\n"%(name))
					f.write("SetValue camera &\".kine.dircns.upvct_active\", False\n")
					# Add bs_tdeLens
					f.write("SIAddArrayElement \"%s.camera.lensshader\"\n"%(name))
					f.write("SIApplyShaderToCnxPoint2 \"Softimage.bs_tdeLens.1.0\", \"%s.camera.Item\"\n"%(name))
					
					#set image aspect ratio
					f.write("SetValue \"%s.%s.aspect\", %.15f\n"%(camera_rig,name,f_aspect))
					#set pixel aspect ratio
					f.write("SetValue \"%s.%s.pixelratio\", %.15f\n"%(camera_rig,name,p_aspect))
		
					#enable filmback config
					f.write("SetValue \"%s.%s.projplane\", True\n"%(camera_rig,name))
					f.write("SetValue \"%s.%s.projplanelockaspect\", False\n"%(camera_rig,name))
					#set filmback width
					f.write("SetValue \"%s.%s.projplanewidth\", %.15f\n"%(camera_rig,name,fback_w))
					f.write("SetValue \"%s.%s.bs_tdeLens.filmBack_W\", %.15f\n"%(camera_rig,name,fback_w))
					#set filmback height
					f.write("SetValue \"%s.%s.projplaneheight\", %.15f\n"%(camera_rig,name,fback_h))
					f.write("SetValue \"%s.%s.bs_tdeLens.filmBack_H\", %.15f\n"%(camera_rig,name,fback_h))

					#set filmbackoffstest 
					f.write("SetValue \"%s.%s.projplaneoffx\", %.15f\n"%(camera_rig,name,prjploffx))
					f.write("SetValue \"%s.%s.projplaneoffy\", %.15f\n"%(camera_rig,name,prjploffy))
					f.write("SetValue \"%s.%s.bs_tdeLens.xOffset\", %.15f\n"%(camera_rig,name,prjploffx))
					f.write("SetValue \"%s.%s.bs_tdeLens.yOffset\", %.15f\n"%(camera_rig,name,prjploffy))

					
					
					p3d	= tde4.getPGroupPosition3D(campg,cam,1)
					p3d	= convertZup(p3d,yup)
					f.write("cam.rotorder = 4\n")
					f.write("Translate \"%s\", %.15f, %.15f, %.15f, siAbsolute\n"%(camera_rig,p3d[0],p3d[1],p3d[2]))
					r3d	= tde4.getPGroupRotation3D(campg,cam,1)
					rot	= convertToAngles(r3d)
					f.write("Rotate \"%s\", %.15f, %.15f, %.15f,siAbsolute \n"%(camera_rig,rot[0],rot[1],rot[2]))

					f.write("SetValue \"%s.%s.kine.local.rotx\", 0\n"%(camera_rig,name))
					f.write("SetValue \"%s.%s.kine.local.roty\", 0\n"%(camera_rig,name))
					f.write("SetValue \"%s.%s.kine.local.rotz\", 0\n"%(camera_rig,name))
					
					# image plane... rotoscopy TODO seqname name.[von..bis;anzahl_channel].endung
					f.write("\n")
					f.write("'' create image plane...\n")
					
					''' Fix Me: The image plane commands causes errors in XSI
					f.write("SetValue \"%s.%s.camdisp.rotoenable\", True\n"%(camera_rig,name))
					f.write("SICreateImageClip \"%s\", \"%s\", clip\n"%(seqname,clipname))
					f.write("SetValue \"%s.%s.rotoscope.imagename\", clip.FullName\n"%(camera_rig,name))
					f.write("SetValue clip & \".timectrl.startoffset\", %d\n"%(1+frame0))
					'''

					# animate camera...
					f.write("\n")
					f.write("'' animating camera %s...\n"%camera_rig)
					f.write("SetValue \"PlayControl.In\", %d\n"%(1+frame0))
					f.write("SetValue \"PlayControl.Out\", %d\n"%(noframes+frame0))
					f.write("\n")
					
					frame	= 1
					
					model = tde4.getLensLDModel(lens)
					
					while frame<=noframes:
						# rot/pos...
						p3d	= tde4.getPGroupPosition3D(campg,cam,frame)
						p3d	= convertZup(p3d,yup)
						r3d	= tde4.getPGroupRotation3D(campg,cam,frame)
						rot	= convertToAngles(r3d)
						if frame>1:
							rot	= [ angleMod360(rot0[0],rot[0]), angleMod360(rot0[1],rot[1]), angleMod360(rot0[2],rot[2]) ]
						rot0	= rot
						f.write("SaveKey \"%s.kine.local.posx\", %d, %.15f \n"%(camera_rig,frame+frame0,p3d[0]))
						f.write("SaveKey \"%s.kine.local.posy\", %d, %.15f \n"%(camera_rig,frame+frame0,p3d[1]))
						f.write("SaveKey \"%s.kine.local.posz\", %d, %.15f \n"%(camera_rig,frame+frame0,p3d[2]))

						f.write("SaveKey \"%s.kine.local.rotx\", %d, %.15f \n"%(camera_rig,frame+frame0,rot[0]))
						f.write("SaveKey \"%s.kine.local.roty\", %d, %.15f \n"%(camera_rig,frame+frame0,rot[1]))
						f.write("SaveKey \"%s.kine.local.rotz\", %d, %.15f \n"%(camera_rig,frame+frame0,rot[2]))

						
						# focal length...
						F	= tde4.getCameraFocalLength(cam,frame)
						focal	= F*10.0
						f.write("SaveKey \"%s.%s.projplanedist\",%d, %.15f\n"%(camera_rig,name,frame+frame0,focal))
						f.write("SaveKey \"%s.%s.bs_tdeLens.focal_length\",%d, %.15f\n"%(camera_rig,name,frame+frame0,focal))
						
						# Set Distortion parameters:
						if model == '3DE Classic LD Model':
							f.write("SaveKey \"%s.%s.bs_tdeLens.model\",%d, %.15f\n"%(camera_rig,name,frame+frame0,0))
							d  = tde4.getLensDistortion(lens,F)
							sq = tde4.getLensAnamorphSqueeze(lens,F)
							xC = tde4.getLensCurvatureX(lens,F)
							yC = tde4.getLensCurvatureY(lens,F)
							q  = tde4.getLensQuarticDistortion(lens,F)
							f.write("SaveKey \"%s.%s.bs_tdeLens.distortion\",%d, %.15f\n"%(camera_rig,name,frame+frame0,d))
							f.write("SaveKey \"%s.%s.bs_tdeLens.anamorphicSqueeze\",%d, %.15f\n"%(camera_rig,name,frame+frame0,sq))
							f.write("SaveKey \"%s.%s.bs_tdeLens.xCurvature\",%d, %.15f\n"%(camera_rig,name,frame+frame0,xC))
							f.write("SaveKey \"%s.%s.bs_tdeLens.yCurvature\",%d, %.15f\n"%(camera_rig,name,frame+frame0,yC))
							f.write("SaveKey \"%s.%s.bs_tdeLens.quarticDist\",%d, %.15f\n"%(camera_rig,name,frame+frame0,q))
						
						if model == '3DE4 Radial - Standard, Degree 4':
							f.write("SaveKey \"%s.%s.bs_tdeLens.model\",%d, %.15f\n"%(camera_rig,name,frame+frame0,1))
							dd2 = tde4.getLensLDAdjustableParameter(lens, "Distortion - Degree 2", F)
							ud2 = tde4.getLensLDAdjustableParameter(lens, "U - Degree 2", F)
							vd2 = tde4.getLensLDAdjustableParameter(lens, "V - Degree 2", F)
							qd4 = tde4.getLensLDAdjustableParameter(lens, "Quartic Distortion - Degree 4", F)
							ud4 = tde4.getLensLDAdjustableParameter(lens, "U - Degree 4", F)
							vd4 = tde4.getLensLDAdjustableParameter(lens, "V - Degree 4", F)
							f.write("SaveKey \"%s.%s.bs_tdeLens.distortion_Degree2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,dd2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.u_Degree2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,ud2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.v_Degree2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,vd2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.quartic_Degree4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,qd4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.u_Degree4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,ud4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.v_Degree4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,vd4))
						
						
						if model == '3DE4 Radial - Fisheye, Degree 8':
							f.write("SaveKey \"%s.%s.bs_tdeLens.model\",%d, %.15f\n"%(camera_rig,name,frame+frame0,2))
							dd2 = tde4.getLensLDAdjustableParameter(lens, "Distortion - Degree 2", F)
							qd4 = tde4.getLensLDAdjustableParameter(lens, "Quartic Distortion - Degree 4", F)
							d6 = tde4.getLensLDAdjustableParameter(lens, "Degree 6", F)
							d8 = tde4.getLensLDAdjustableParameter(lens, "Degree 8", F)
							f.write("SaveKey \"%s.%s.bs_tdeLens.quartic_Degree4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,qd4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.Degree6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.Degree8\",%d, %.15f\n"%(camera_rig,name,frame+frame0,d8))
							
						if model == '3DE4 Anamorphic, Degree 6':
							f.write("SaveKey \"%s.%s.bs_tdeLens.model\",%d, %.15f\n"%(camera_rig,name,frame+frame0,3))
							cx02d2 = tde4.getLensLDAdjustableParameter(lens, "Cx02 - Degree 2", F)
							cy02d2 = tde4.getLensLDAdjustableParameter(lens, "Cy02 - Degree 2", F)
							cx22d2 = tde4.getLensLDAdjustableParameter(lens, "Cx22 - Degree 2", F)
							cy22d2 = tde4.getLensLDAdjustableParameter(lens, "Cy22 - Degree 2", F)
							cx04d4 = tde4.getLensLDAdjustableParameter(lens, "Cx04 - Degree 4", F)
							cy04d4 = tde4.getLensLDAdjustableParameter(lens, "Cy04 - Degree 4", F)
							cx24d4 = tde4.getLensLDAdjustableParameter(lens, "Cx24 - Degree 4", F)
							cy24d4 = tde4.getLensLDAdjustableParameter(lens, "Cy24 - Degree 4", F)
							cx44d4 = tde4.getLensLDAdjustableParameter(lens, "Cx44 - Degree 4", F)
							cy44d4 = tde4.getLensLDAdjustableParameter(lens, "Cy44 - Degree 4", F)
							cx06d6 = tde4.getLensLDAdjustableParameter(lens, "Cx06 - Degree 6", F)
							cy06d6 = tde4.getLensLDAdjustableParameter(lens, "Cy06 - Degree 6", F)
							cx26d6 = tde4.getLensLDAdjustableParameter(lens, "Cx26 - Degree 6", F)
							cy26d6 = tde4.getLensLDAdjustableParameter(lens, "Cy26 - Degree 6", F)
							cx46d6 = tde4.getLensLDAdjustableParameter(lens, "Cx46 - Degree 6", F)
							cy46d6 = tde4.getLensLDAdjustableParameter(lens, "Cy46 - Degree 6", F)
							cx66d6 = tde4.getLensLDAdjustableParameter(lens, "Cx66 - Degree 6", F)
							cy66d6 = tde4.getLensLDAdjustableParameter(lens, "Cy66 - Degree 6", F)
							
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx02d2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx02d2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy02d2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy02d2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx22d2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx22d2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy22d2\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy22d2))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx04d4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx04d4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy04d4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy04d4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx24d4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx24d4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy24d4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy24d4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx44d4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx44d4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy44d4\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy44d4))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx06d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx06d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy06d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy06d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx26d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx26d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy26d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy26d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx46d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx46d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy46d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy46d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cx66d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cx66d6))
							f.write("SaveKey \"%s.%s.bs_tdeLens.cy66d6\",%d, %.15f\n"%(camera_rig,name,frame+frame0,cy66d6))

						
						frame	+= 1
				
			#
			# write camera point group...
			
			f.write("\n")
			f.write("'' create camera point group...\n")
			name	= "cameraPGroup_%s_1"%validName(tde4.getPGroupName(campg))
			f.write("set pointGroup = sceneGroupName.AddNull(\"%s\")\n"%name)
			f.write("SetValue pointGroup & \".null.primary_icon\", 0\n")
			
			f.write("\n")
			
			# write points...
			l	= tde4.getPointList(campg)
			for p in l:
				if tde4.isPointCalculated3D(campg,p):
					name	= tde4.getPointName(campg,p)
					name	= "p%s"%validName(name)
					p3d	= tde4.getPointCalcPosition3D(campg,p)
					p3d	= convertZup(p3d,yup)
					f.write("\n")
					f.write("'' create point %s...\n"%name)
					f.write("set pt = pointGroup.AddNull(\"%s\")\n"%name)
					f.write("Translate pt,%.15f,%.15f,%.15f,siAbsolute\n"%(p3d[0],p3d[1],p3d[2]))
			
			f.write("\n")
			
			
			#
			# write object point groups...
			
			camera		= tde4.getCurrentCamera()
			noframes	= tde4.getCameraNoFrames(camera)
			pgl		= tde4.getPGroupList()
			index		= 1
			for pg in pgl:
				if tde4.getPGroupType(pg)=="OBJECT" and camera!=None:
					f.write("\n")
					f.write("'' create object point group...\n")
					pgname	= "objectPGroup_%s_%d_1"%(validName(tde4.getPGroupName(pg)),index)
					index	+= 1
					f.write("set pointGroup = sceneGroupName.AddNull(\"%s\")\n"%pgname)
					f.write("SetValue pointGroup & \".null.primary_icon\", 0\n")
					f.write("SetValue pointGroup & \".kine.local.rotorder\", 4\n")
					
					# write points...
					l	= tde4.getPointList(pg)
					for p in l:
						if tde4.isPointCalculated3D(pg,p):
							name	= tde4.getPointName(pg,p)
							name	= "p%s"%validName(name)
							p3d	= tde4.getPointCalcPosition3D(pg,p)
							p3d	= convertZup(p3d,yup)
							f.write("\n")
							f.write("'' create point %s...\n"%name)
							f.write("set pt = pointGroup.AddNull(\"%s\")\n"%name)
							f.write("Translate pt,%.15f,%.15f,%.15f,siAbsolute\n"%(p3d[0],p3d[1],p3d[2]))
					
					f.write("\n")
					scale	= tde4.getPGroupScale3D(pg)
					f.write("Scale pointGroup, %.15f, %.15f, %.15f \n"%(scale,scale,scale))

					# animate object point group...
					f.write("\n")
					f.write("'' animating point group %s...\n"%pgname)
					frame	= 1
					while frame<=noframes:
						# rot/pos...
						p3d	= tde4.getPGroupPosition3D(pg,camera,frame)
						p3d	= convertZup(p3d,yup)
						r3d	= tde4.getPGroupRotation3D(pg,camera,frame)
						rot	= convertToAngles(r3d)
						if frame>1:
							rot	= [ angleMod360(rot0[0],rot[0]), angleMod360(rot0[1],rot[1]), angleMod360(rot0[2],rot[2]) ]
						rot0	= rot
						f.write("SaveKey pointGroup & \".kine.local.posx\", %d, %.15f \n"%(frame+frame0,p3d[0]))
						f.write("SaveKey pointGroup & \".kine.local.posy\", %d, %.15f \n"%(frame+frame0,p3d[1]))
						f.write("SaveKey pointGroup & \".kine.local.posz\", %d, %.15f \n"%(frame+frame0,p3d[2]))

						f.write("SaveKey pointGroup & \".kine.local.rotx\", %d, %.15f \n"%(frame+frame0,rot[0]))
						f.write("SaveKey pointGroup & \".kine.local.roty\", %d, %.15f \n"%(frame+frame0,rot[1]))
						f.write("SaveKey pointGroup & \".kine.local.rotz\", %d, %.15f \n"%(frame+frame0,rot[2]))

						frame	+= 1
			
		
			#
			# global (scene node) transformation...
			
			p3d	= tde4.getScenePosition3D()
			p3d	= convertZup(p3d,yup)
			r3d	= tde4.getSceneRotation3D()
			rot	= convertToAngles(r3d)
			s	= tde4.getSceneScale3D()
			f.write("Translate sceneGroupName , %.15f, %.15f, %.15f, siAbsolute\n"%(p3d[0],p3d[1],p3d[2])) 
			f.write("Rotate sceneGroupName , %.15f, %.15f, %.15f, siAbsolute\n"%(rot[0],rot[1],rot[2]))
			f.write("Scale sceneGroupName , %.15f, %.15f, %.15f, siAbsolute\n"%(s,s,s))

						
			f.write("\n")
			
			f.write("'' Switch command logging on again\n")
			f.write("SetValue \"preferences.scripting.cmdlog\", False\n")

			f.close()
			tde4.postQuestionRequester("Export Softimage...","Project successfully exported.","Ok")
			
		else:
			tde4.postQuestionRequester("Export Softimage...","Error, couldn't open file.","Ok")


tde4.deleteCustomRequester(req)


